"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _objectSpread2 = _interopRequireDefault(require("@babel/runtime/helpers/objectSpread"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _set2 = _interopRequireDefault(require("lodash/set"));

var _isObject2 = _interopRequireDefault(require("lodash/isObject"));

var _cloneDeep2 = _interopRequireDefault(require("lodash/cloneDeep"));

var _merge2 = _interopRequireDefault(require("lodash/merge"));

var _get2 = _interopRequireDefault(require("lodash/get"));

var _webinyModel = require("webiny-model");

var _driver = _interopRequireDefault(require("./driver"));

var _entityPool = _interopRequireDefault(require("./entityPool"));

var _createPaginationMeta = _interopRequireDefault(require("./createPaginationMeta"));

var _eventHandler = _interopRequireDefault(require("./eventHandler"));

var _entityCollection = _interopRequireDefault(require("./entityCollection"));

var _entityModel = _interopRequireDefault(require("./entityModel"));

var _entityAttributesContainer = _interopRequireDefault(require("./entityAttributesContainer"));

var _queryResult = _interopRequireDefault(require("./queryResult"));

var _index = require("./index");

class Entity {
  constructor() {
    (0, _defineProperty2.default)(this, "classId", void 0);
    (0, _defineProperty2.default)(this, "storageClassId", void 0);
    (0, _defineProperty2.default)(this, "id", void 0);
    (0, _defineProperty2.default)(this, "model", void 0);
    (0, _defineProperty2.default)(this, "listeners", void 0);
    (0, _defineProperty2.default)(this, "existing", void 0);
    (0, _defineProperty2.default)(this, "processing", void 0);
    (0, _defineProperty2.default)(this, "createdOn", void 0);
    (0, _defineProperty2.default)(this, "updatedOn", void 0);
    (0, _defineProperty2.default)(this, "savedOn", void 0);
    (0, _defineProperty2.default)(this, "deleted", void 0);
    const proxy = new Proxy(this, {
      set: (instance, key, value) => {
        const attribute = instance.getModel().getAttribute(key);

        if (attribute) {
          attribute.setValue(value);
          return true;
        }

        instance[key] = value;
        return true;
      },
      get: (instance, key) => {
        if (["classId", "storageClassId", "driver"].includes(key)) {
          return instance.constructor[key];
        }

        const attribute = instance.getModel().getAttribute(key);

        if (attribute) {
          return attribute.getValue();
        }

        return instance[key];
      }
    });
    const modelClass = this.getDriver().getModelClass();

    if (!modelClass) {
      throw new _index.EntityError("Entity model is missing.", _index.EntityError.MODEL_MISSING);
    }

    this.model = new modelClass().setParentEntity(proxy);
    this.listeners = {};
    this.existing = false;
    this.processing = null;

    if ((0, _get2.default)(this, "constructor.crud.logs")) {
      this.attr("savedOn").date().setSkipOnPopulate();
      this.attr("createdOn").date().setSkipOnPopulate();
      this.attr("updatedOn").date().setSkipOnPopulate();
    }

    const deletes = (0, _get2.default)(this, "constructor.crud.delete", {});

    if (deletes.soft) {
      this.attr("deleted").boolean().setSkipOnPopulate().setValue(false);
      this.on("beforeDelete", () => proxy.deleted = true);
    }

    this.on("delete", () => {
      if (!proxy.id) {
        throw new _index.EntityError("Entity cannot be deleted because it was not previously saved.", _index.EntityError.CANNOT_DELETE_NO_ID);
      }
    });
    this.getDriver().onEntityConstruct(proxy);

    if (!this.getAttribute("id")) {
      this.attr("id").char();
    }

    return proxy;
  }
  /**
   * Returns instance of entity's model.
   */


  getModel() {
    return this.model;
  }
  /**
   * Returns instance of used driver.
   */


  static getDriver() {
    return this.driver;
  }
  /**
   * Returns instance of used driver.
   */


  getDriver() {
    return this.constructor.driver;
  }
  /**
   * Returns instance of used entity pool.
   */


  static getEntityPool() {
    return this.pool;
  }
  /**
   * Returns instance of used entity pool.
   */


  getEntityPool() {
    return this.constructor.pool;
  }

  static isEntityInstance(value) {
    return !!(0, _get2.default)(value, "constructor.classId");
  }

  static isEntityClass(value) {
    return !!(0, _get2.default)(value, "classId");
  }

  static isInstanceOf(instance, instanceClass) {
    return (0, _get2.default)(instance, "constructor.classId") === (0, _get2.default)(instanceClass, "classId");
  }
  /**
   * Sets whether entity is existing or not.
   */


  setExisting(flag = true) {
    this.existing = flag;
    return this;
  }
  /**
   * Returns true if entity exists or in other words, is already saved in storage. Otherwise returns false.
   */


  isExisting() {
    return this.existing;
  }
  /**
   * Creates new attribute with name.
   */


  attr(name) {
    const container = this.getModel().getAttributesContainer();
    return container.attr(name);
  }
  /**
   * Returns single attribute by given name.
   */


  getAttribute(name) {
    return this.getModel().getAttribute(name);
  }
  /**
   * Returns all entity's attributes.
   */


  getAttributes() {
    return this.getModel().getAttributes();
  }

  async get(path = "", defaultValue) {
    return this.getModel().get(path, defaultValue);
  }

  async set(path, value) {
    return this.getModel().set(path, value);
  }
  /**
   * Returns entity's JSON representation.
   */


  async toJSON(path) {
    const attribute = this.getAttribute("id");
    return (0, _merge2.default)({
      id: attribute.getValue()
    }, path ? await this.getModel().toJSON(path) : {});
  }
  /**
   * Returns data suitable for storage.
   */


  async toStorage() {
    return this.getModel().toStorage();
  }
  /**
   * Validates current entity and throws exception that contains all invalid attributes.
   */


  async validate() {
    await this.emit("beforeValidate");
    await this.getModel().validate();
    await this.emit("afterValidate");
  }
  /**
   * Used to populate entity with given data.
   */


  populate(data) {
    this.getModel().populate(data);
    return this;
  }
  /**
   * Used when populating entity with data from storage.
   * @param data
   */


  populateFromStorage(data) {
    this.getModel().populateFromStorage(data);
    return this;
  }
  /**
   * Returns class name.
   */


  getClassName() {
    return this.constructor.name;
  }
  /**
   * Returns class name.
   */


  static getClassName() {
    return this.name;
  }
  /**
   * Tells us whether a given ID is valid or not.
   * @param id
   * @param params
   */


  isId(id, params = {}) {
    return this.getDriver().isId(this, id, (0, _cloneDeep2.default)(params));
  }
  /**
   * Tells us whether a given ID is valid or not.
   * @param id
   * @param params
   */


  static isId(id, params = {}) {
    return this.getDriver().isId(this, id, (0, _cloneDeep2.default)(params));
  }
  /**
   * Saves current and all linked entities (if autoSave on the attribute was enabled).
   * @param params
   */


  async save(params) {
    if (!params) {
      params = {};
    }

    const events = params.events || {};
    const existing = this.isExisting();

    if (this.processing) {
      return;
    }

    this.processing = "save";

    if (existing) {
      events.beforeUpdate !== false && (await this.emit("beforeUpdate", {
        params
      }));
    } else {
      events.beforeCreate !== false && (await this.emit("beforeCreate", {
        params
      }));
    }

    events.beforeSave !== false && (await this.emit("beforeSave", {
      params
    }));
    const logs = (0, _get2.default)(this, "constructor.crud.logs");

    try {
      if (existing) {
        events.update !== false && (await this.emit("__update", {
          params
        }));
      } else {
        events.create !== false && (await this.emit("__create", {
          params
        }));
      }

      events.save !== false && (await this.emit("__save", {
        params
      }));
      params.validation !== false && (await this.validate());

      if (existing) {
        events.__beforeUpdate !== false && (await this.emit("__beforeUpdate", {
          params
        }));

        if (logs && this.isDirty()) {
          this.savedOn = new Date();
          this.updatedOn = new Date();
        }
      } else {
        events.__beforeCreate !== false && (await this.emit("__beforeCreate", {
          params
        }));

        if (logs && this.isDirty()) {
          this.savedOn = new Date();
          this.createdOn = new Date();
        }
      }

      events.__beforeSave !== false && (await this.emit("__beforeSave", {
        params
      }));

      if (this.isDirty()) {
        await this.getDriver().save(this, params);
      }

      if (existing) {
        events.__afterUpdate !== false && (await this.emit("__afterUpdate", {
          params
        }));
      } else {
        events.__afterCreate !== false && (await this.emit("__afterCreate", {
          params
        }));
      }

      events.__afterSave !== false && (await this.emit("__afterSave", {
        params
      }));
      this.setExisting();
      this.getModel().clean();
      this.getEntityPool().add(this);
    } catch (e) {
      throw e;
    } finally {
      this.processing = null;
    }

    if (existing) {
      events.afterUpdate !== false && (await this.emit("afterUpdate", {
        params
      }));
    } else {
      events.afterCreate !== false && (await this.emit("afterCreate", {
        params
      }));
    }

    events.afterSave !== false && (await this.emit("afterSave", {
      params
    }));
  }
  /**
   * Deletes current and all linked entities (if autoDelete on the attribute was enabled).
   * @param params
   */


  async delete(params) {
    if (this.processing) {
      return;
    }

    this.processing = "delete";
    const soft = (0, _get2.default)(this, "constructor.crud.delete.soft");

    if (!params) {
      params = {};
    }

    try {
      const events = params.events || {};
      events.delete !== false && (await this.emit("delete", {
        params
      }));
      params.validation !== false && (await this.validate());
      events.beforeDelete !== false && (await this.emit("beforeDelete", {
        params
      }));
      const logs = (0, _get2.default)(this, "constructor.crud.logs");

      if (logs) {
        this.savedOn = new Date();
        this.updatedOn = new Date();
      }

      if (soft && params.permanent !== true) {
        await this.getDriver().save(this, params);
      } else {
        await this.getDriver().delete(this, params);
      }

      events.afterDelete !== false && (await this.emit("afterDelete", {
        params
      }));
      this.getEntityPool().remove(this);
    } catch (e) {
      throw e;
    } finally {
      this.processing = null;
    }
  }
  /**
   * Finds a single entity matched by given ID.
   * @param id
   * @param params
   */


  static async findById(id, params) {
    if (!id) {
      return null;
    }

    const pooled = this.getEntityPool().get(this, id);

    if (pooled) {
      return pooled;
    }

    if (!params) {
      params = {};
    }

    const newParams = (0, _merge2.default)((0, _cloneDeep2.default)(params), {
      query: {
        id
      }
    });
    return await this.findOne(newParams);
  }
  /**
   * Finds one or more entities matched by given IDs.
   * @param ids
   * @param params
   */


  static async findByIds(ids, params) {
    const output = [];

    for (let i = 0; i < ids.length; i++) {
      const entity = await this.findById(ids[i], params);

      if (entity) {
        output.push(entity);
      }
    }

    return output;
  }
  /**
   * Finds one entity matched by given query parameters.
   * @param params
   */


  static async findOne(params) {
    if (!params) {
      params = {};
    }

    const prepared = this.__prepareParams(params);

    await this.emit("query", prepared);
    const queryResult = await this.getDriver().findOne(this, prepared);
    const result = queryResult.getResult();

    if ((0, _isObject2.default)(result)) {
      const pooled = this.getEntityPool().get(this, result.id);

      if (pooled) {
        await pooled.emit("read", {
          params
        });
        return pooled;
      }

      const entity = new this().setExisting().populateFromStorage(result);
      await entity.emit("read", {
        params
      });
      this.getEntityPool().add(entity);
      return entity;
    }

    return null;
  }
  /**
   * Finds one or more entities matched by given query parameters.
   * @param params
   */


  static async find(params) {
    if (!params) {
      params = {};
    }

    const prepared = this.__prepareParams(params); // Prepare find-specific params: perPage and page.


    prepared.page = Number(prepared.page);

    if (!Number.isInteger(prepared.page) || prepared.page <= 0) {
      prepared.page = 1;
    }

    prepared.perPage = Number(prepared.perPage);

    if (Number.isInteger(prepared.perPage) && prepared.perPage > 0) {
      const maxPerPage = (0, _get2.default)(this, "crud.read.maxPerPage");

      if (Number.isInteger(maxPerPage) && prepared.perPage > maxPerPage) {
        throw new _index.EntityError(`Cannot query for more than ${maxPerPage} entities per page.`, _index.EntityError.MAX_PER_PAGE_EXCEEDED);
      }
    } else {
      prepared.perPage = 10;
    }

    await this.emit("query", prepared);
    const queryResult = await this.getDriver().find(this, prepared);
    const entityCollection = new _entityCollection.default().setParams(prepared).setMeta((0, _objectSpread2.default)({}, (0, _createPaginationMeta.default)(), queryResult.getMeta()));
    const result = queryResult.getResult();

    if (result instanceof Array) {
      for (let i = 0; i < result.length; i++) {
        const pooled = this.getEntityPool().get(this, result[i].id);

        if (pooled) {
          entityCollection.push(pooled);
          await pooled.emit("read", {
            params
          });
        } else {
          const entity = new this().setExisting().populateFromStorage(result[i]);
          await entity.emit("read", {
            params
          });
          this.getEntityPool().add(entity);
          entityCollection.push(entity);
        }
      }
    }

    return entityCollection;
  }
  /**
   * Counts total number of entities matched by given query parameters.
   * @param params
   */


  static async count(params) {
    if (!params) {
      params = {};
    }

    const prepared = this.__prepareParams(params);

    await this.emit("query", prepared);
    const queryResult = await this.getDriver().count(this, prepared);
    return queryResult.getResult();
  }
  /**
   * Registers a listener that will be triggered only on current entity instance.
   * @param name
   * @param callback
   */


  on(name, callback) {
    const eventHandler = new _eventHandler.default(name, callback);

    if (!this.listeners[eventHandler.getName()]) {
      this.listeners[eventHandler.getName()] = [];
    }

    this.listeners[eventHandler.getName()].push(eventHandler);
    return eventHandler;
  }
  /**
   * Registers a listener that will be triggered on all entity instances of current class.
   * @param name
   * @param callback
   */


  static on(name, callback) {
    if (!this.listeners) {
      this.listeners = {};
    }

    const eventHandler = new _eventHandler.default(name, callback);

    if (!this.listeners[eventHandler.getName()]) {
      this.listeners[eventHandler.getName()] = [];
    }

    this.listeners[eventHandler.getName()].push(eventHandler);
    return eventHandler;
  }
  /**
   * Emits an event, which will trigger both static and instance listeners.
   * @param name
   * @param data
   */


  async emit(name, data = {}) {
    if (this.listeners[name]) {
      for (let i = 0; i < this.listeners[name].length; i++) {
        await this.listeners[name][i].execute((0, _objectSpread2.default)({}, data, {
          entity: this
        }));
      }
    }

    if (this.constructor.listeners) {
      if (this.constructor.listeners[name]) {
        for (let i = 0; i < this.constructor.listeners[name].length; i++) {
          await this.constructor.listeners[name][i].execute((0, _objectSpread2.default)({}, data, {
            entity: this
          }));
        }
      }
    }

    return this;
  }
  /**
   * Emits an event, which will trigger static event listeners.
   * @param name
   * @param data
   */


  static async emit(name, data = {}) {
    if ((0, _get2.default)(this, "listeners." + name)) {
      for (let i = 0; i < this.listeners[name].length; i++) {
        await this.listeners[name][i].execute((0, _objectSpread2.default)({}, data, {
          entity: this
        }));
      }
    }
  }

  isDirty() {
    return this.getModel().isDirty();
  }

  isClean() {
    return !this.isDirty();
  }
  /**
   * Returns information about the entity.
   * @returns {{name: string, id: *, attributes: {name: *, class: string|string|string|string|string|string|string}[]}}
   */


  static describe(options = {}) {
    const instance = new this();
    const omit = (0, _get2.default)(options, "attributes.omit", []);
    return {
      name: instance.getClassName(),
      classId: (0, _get2.default)(instance, "classId"),
      attributes: Object.keys(instance.getAttributes()).map(key => {
        const attribute = instance.getAttribute(key);

        if (!attribute) {
          return null;
        }

        if (omit.includes(key)) {
          return null;
        }

        return {
          name: attribute.getName(),
          protected: attribute.getSkipOnPopulate(),
          class: attribute.constructor.name
        };
      }).filter(value => value)
    };
  }
  /**
   * Creates a clone of given params first.
   * If soft delete functionality is enabled, this will append "deleted = false" filter, meaning only non-deleted
   * entities can be taken into consideration. If "deleted" flag was already set from the outside, it won't take any action.
   * @param params
   * @private
   */


  static __prepareParams(params) {
    const clone = (0, _cloneDeep2.default)(params);

    if ((0, _get2.default)(this, "crud.delete.soft") === true) {
      (0, _set2.default)(clone, "query.deleted", (0, _get2.default)(clone, "query.deleted", false));
    }

    return clone;
  }

}

(0, _defineProperty2.default)(Entity, "classId", void 0);
(0, _defineProperty2.default)(Entity, "storageClassId", void 0);
(0, _defineProperty2.default)(Entity, "driver", void 0);
(0, _defineProperty2.default)(Entity, "pool", void 0);
(0, _defineProperty2.default)(Entity, "listeners", void 0);
(0, _defineProperty2.default)(Entity, "crud", void 0);
Entity.classId = "";
Entity.storageClassId = "";
Entity.driver = new _driver.default();
Entity.pool = new _entityPool.default();
Entity.crud = {
  logs: false,
  delete: {
    soft: false
  },
  read: {
    maxPerPage: 100
  }
};
var _default = Entity;
exports.default = _default;
//# sourceMappingURL=entity.js.map