"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _get2 = _interopRequireDefault(require("lodash/get"));

var _webinyModel = require("webiny-model");

var _entity = _interopRequireDefault(require("./../entity"));

var _entityCollection = _interopRequireDefault(require("./../entityCollection"));

class EntitiesAttributeValue extends _webinyModel.AttributeValue {
  constructor(attribute) {
    super(attribute);
    (0, _defineProperty2.default)(this, "initial", void 0);
    (0, _defineProperty2.default)(this, "links", void 0);
    (0, _defineProperty2.default)(this, "queue", void 0);
    (0, _defineProperty2.default)(this, "parentModel", void 0);
    this.current = new _entityCollection.default();
    this.initial = new _entityCollection.default();
    this.links = {
      dirty: false,
      set: false,
      current: new _entityCollection.default(),
      initial: new _entityCollection.default()
    };
    this.state = {
      loading: false,
      loaded: false
    };
    this.queue = [];
    this.parentModel = this.attribute.getParentModel();
  }
  /**
   * Ensures data is loaded correctly, and in the end returns current value.
   * @returns {Promise<*>}
   */


  async load() {
    if (this.isLoading()) {
      return new Promise(resolve => this.queue.push(resolve));
    }

    if (this.isLoaded()) {
      return;
    }

    const attribute = this.attribute;
    const classes = attribute.classes;
    this.state.loading = true;

    if (this.parentModel.getParentEntity().isExisting()) {
      const idAttribute = this.parentModel.getParentEntity().getAttribute("id");
      let id = await idAttribute.getStorageValue();

      if (classes.using.class) {
        this.links.initial = await classes.using.class.find({
          query: {
            [classes.entities.attribute]: id
          }
        });
        this.initial = new _entityCollection.default();

        for (let i = 0; i < this.links.initial.length; i++) {
          this.initial.push((await this.links.initial[i][classes.using.attribute]));
        }
      } else {
        this.initial = await classes.entities.class.find({
          query: {
            [classes.entities.attribute]: id
          }
        });
      }

      if (this.isClean()) {
        const initial = this.getInitial();
        const initialLinks = this.getInitialLinks();

        if (Array.isArray(initial) && Array.isArray(initialLinks)) {
          this.setCurrent(new _entityCollection.default(initial), {
            skipDifferenceCheck: true
          });

          if (classes.using.class) {
            this.setCurrentLinks(new _entityCollection.default(initialLinks), {
              skipDifferenceCheck: true
            });
          }
        }
      }
    }

    this.state.loading = false;
    this.state.loaded = true;
    await this.__executeQueue();
    return this.current;
  }

  setInitial(value) {
    this.initial = value;
    return this;
  }

  getInitial() {
    return this.initial;
  }

  hasInitial() {
    return this.getInitial().length > 0;
  }

  hasCurrent() {
    return this.getCurrent().length > 0;
  }

  async deleteInitial() {
    // If initial is empty, that means nothing was ever loaded (attribute was not accessed) and there is nothing to do.
    // Otherwise, deleteInitial method will internally delete only entities that are not needed anymore.
    if (!this.hasInitial()) {
      return;
    }

    const initial = this.getInitial(),
          currentEntitiesIds = this.getCurrent().map(entity => (0, _get2.default)(entity, "id"));

    for (let i = 0; i < initial.length; i++) {
      const currentInitial = initial[i];

      if (_entity.default.isEntityInstance(currentInitial)) {
        if (!currentEntitiesIds.includes(currentInitial.id)) {
          await currentInitial.delete();
        }
      }
    }
  }
  /**
   * Creates a new array that contains all currently loaded entities.
   */


  syncInitial() {
    this.initial = new _entityCollection.default(this.getCurrent().map(entity => entity));
  }

  async manageCurrent() {
    const current = this.getCurrent();

    for (let i = 0; i < current.length; i++) {
      const entity = current[i];

      if (_entity.default.isEntityInstance(entity)) {
        const attribute = this.attribute;
        const classes = attribute.classes;
        await entity.set(classes.entities.attribute, this.parentModel.getParentEntity());
      }
    }
  }

  getCurrent() {
    return super.getCurrent();
  }

  getInitialLinks() {
    return this.links.initial;
  }

  hasInitialLinks() {
    return this.getInitialLinks().length > 0;
  }

  setInitialLinks(value) {
    this.links.initial = value;
    return this;
  }

  getCurrentLinks() {
    return this.links.current;
  }

  hasCurrentLinks() {
    return this.getCurrentLinks().length > 0;
  }

  setCurrentLinks(value, options = {}) {
    this.links.set = true;

    if (!options.skipDifferenceCheck) {
      if (this.isDifferentFrom(value)) {
        this.links.dirty = true;
      }
    }

    this.links.current = value;
    return this;
  }

  async deleteInitialLinks() {
    // If initial is empty, that means nothing was ever loaded (attribute was not accessed) and there is nothing to do.
    // Otherwise, deleteInitial method will internally delete only entities that are not needed anymore.
    if (!this.hasInitialLinks()) {
      return;
    }

    const initialLinks = this.getInitialLinks(),
          // $FlowFixMe
    currentLinksIds = this.getCurrentLinks().map(entity => entity.id);

    for (let i = 0; i < initialLinks.length; i++) {
      const initial = initialLinks[i]; // $FlowFixMe

      if (!currentLinksIds.includes(initial.id)) {
        _entity.default.isEntityInstance(initial) && (await initial.delete());
      }
    }
  }
  /**
   * Creates a new array that contains all currently loaded entities.
   */


  syncInitialLinks() {
    this.links.initial = this.getCurrentLinks().map(entity => entity);
  }
  /**
   * Sets current links, based on initial and currently set entities.
   *
   * How links-management works?
   * When entities are set, on "__save" event, attribute will be first loaded - meaning all initial (from storage)
   * linked entities and its links will be loaded ("this.initial" / "this.links.initial"). After that, this method
   * will iterate over all newly set entities, and check if for each a link is already existing. If so, it will
   * use it, otherwise a new instance is created, linking parent and set entity together.
   *
   * Additional note: previously, there was an idea that link entities could also contain additional information.
   * This still could works for lists in which entities are all unique, meaning all entities show only once in
   * the list. In cases where a single entity can appear more than once, this might not be the best solution, since
   * linking problems can appear.
   *
   * Eg. if user has a list of entities: A - A - B - C, and if links have a specific information, reordering
   * first two A entities wouldn't make a difference, and nothing would be updated.
   *
   * But generally, this is a bad approach to have, in cases where links need to have additional data, a new entity
   * would have to be made, linking the A product and containing all needed information.
   *
   * Basic example of this is a cart, with added products. Added product might appear many times, in different
   * colors and sizes, so here it would be best to just create CartItem entity, that links the product and contains
   * needed information.
   *
   * Link entities can be extended with additional attributes where it's sure that no duplicates can occur.
   * @returns {Promise<void>}
   */


  async manageCurrentLinks() {
    const links = [],
          current = this.getCurrent(),
          currentLinks = this.getInitialLinks();

    for (let i = 0; i < current.length; i++) {
      const currentEntity = current[i]; // Following chunk actually represents: "_.find(currentLinks, link => link.<entity> === current);".
      // "for" loop used because of async operations.

      let link = null;

      for (let j = 0; j < currentLinks.length; j++) {
        // $FlowFixMe
        const linkedEntity = await currentLinks[j][this.attribute.getUsingAttribute()];

        if (linkedEntity === currentEntity) {
          link = currentLinks[j];
          break;
        }
      } // If entity has an already existing link instance, it will be used. Otherwise a new instance will be created.
      // Links array cannot contain two same instances.


      if (link && !links.includes(link)) {
        links.push(link);
      } else {
        const attr = {
          instance: this.attribute
        };
        attr.usingClass = attr.instance.getUsingClass();
        attr.usingAttribute = attr.instance.getUsingAttribute();
        attr.entitiesAttribute = attr.instance.getEntitiesAttribute();

        if (attr.usingClass && attr.usingAttribute && attr.entitiesAttribute) {
          const entity = new attr.usingClass();
          await entity.set(attr.usingAttribute, currentEntity);
          await entity.set(attr.entitiesAttribute, attr.instance.getParentModel().getParentEntity());
          links.push(entity);
        }
      }
    }

    this.setCurrentLinks(links);
  }
  /**
   * Value cannot be set as clean if ID is missing in one of the entities.
   * @returns {this}
   */


  clean() {
    const current = this.getCurrent();

    for (let i = 0; i < current.length; i++) {
      if (_entity.default.isEntityInstance(current[i])) {
        if (!current[i].id) {
          return this;
        }
      }
    }

    super.clean();
    return this;
  }

  isDirty() {
    if (super.isDirty()) {
      return true;
    }

    if (Array.isArray(this.current)) {
      for (let i = 0; i < this.current.length; i++) {
        if (_entity.default.isEntityInstance(this.current[i]) && this.current[i].isDirty()) {
          return true;
        }
      }
    }

    return false;
  }

  isClean() {
    return !this.isDirty();
  }

  async __executeQueue() {
    if (this.queue.length) {
      for (let i = 0; i < this.queue.length; i++) {
        await this.queue[i]();
      }

      this.queue = [];
    }
  }

}

var _default = EntitiesAttributeValue;
exports.default = _default;
//# sourceMappingURL=entitiesAttributeValue.js.map