"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _get2 = _interopRequireDefault(require("lodash/get"));

var _webinyModel = require("webiny-model");

var _index = require("../index");

class EntityAttributeValue extends _webinyModel.AttributeValue {
  constructor(attribute) {
    super(attribute);
    (0, _defineProperty2.default)(this, "queue", void 0);
    (0, _defineProperty2.default)(this, "initial", void 0);
    (0, _defineProperty2.default)(this, "parentModel", void 0);
    this.queue = []; // Contains initial value received upon loading from storage. If the current value becomes different from initial,
    // upon save, old entity must be removed. This is only active when auto delete option on the attribute is enabled,
    // which then represents a one to one relationship.

    this.initial = null;
    this.parentModel = this.attribute.getParentModel();
  }
  /**
   * Ensures data is loaded correctly, and in the end returns current value.
   * @returns {Promise<*>}
   */


  async load() {
    if (this.isLoading()) {
      return new Promise(resolve => {
        this.queue.push(resolve);
      });
    }

    if (this.isLoaded()) {
      return;
    }

    this.state.loading = true; // Only if we have a valid ID set, we must load linked entity.

    const initial = this.getInitial();

    if (this.parentModel.getParentEntity().isId(initial)) {
      const attribute = this.attribute;
      const entityClass = attribute.getEntityClass();

      if (entityClass) {
        const entity = await entityClass.findById(initial);
        this.setInitial(entity); // If current value is not dirty, than we can set initial value as current, otherwise we
        // assume that something else was set as current value like a new entity.

        if (this.isClean()) {
          this.setCurrent(entity, {
            skipDifferenceCheck: true
          });
        }
      }
    }

    this.state.loading = false;
    this.state.loaded = true;
    await this.__executeQueue();
    return this.getCurrent();
  }

  async deleteInitial(options) {
    if (!this.hasInitial()) {
      return;
    } // Initial value will always be an existing (already saved) Entity instance.


    const initial = this.getInitial();

    if (_index.Entity.isEntityInstance(initial) && (0, _get2.default)(initial, "id") !== (0, _get2.default)(this.getCurrent(), "id")) {
      await initial.delete(options);
    }
  }

  syncInitial() {
    this.setInitial(this.getCurrent());
  }

  setInitial(value) {
    this.initial = value;
    return this;
  }

  getInitial() {
    return this.initial;
  }

  hasInitial() {
    const attribute = this.attribute;
    return _index.Entity.isInstanceOf(this.initial, attribute.getEntityClass());
  }

  isDirty() {
    if (super.isDirty()) {
      return true;
    }

    return _index.Entity.isEntityInstance(this.current) && this.current.isDirty();
  }

  isClean() {
    return !this.isDirty();
  }
  /**
   * Value cannot be set as clean if there is no ID present.
   * @returns {EntityAttributeValue}
   */


  clean() {
    if ((0, _get2.default)(this.getCurrent(), "id")) {
      super.clean();
    }

    return this;
  }

  isDifferentFrom(value) {
    const currentId = (0, _get2.default)(this.current, "id", this.current);

    if (_index.Entity.isEntityInstance(value)) {
      return !value.id || value.id !== currentId;
    }

    if (value instanceof Object) {
      if (!value.id) {
        return true;
      }

      return value.id !== currentId || Object.keys(value).length > 1;
    }

    return currentId !== value;
  }

  async __executeQueue() {
    if (this.queue.length) {
      for (let i = 0; i < this.queue.length; i++) {
        await this.queue[i]();
      }

      this.queue = [];
    }
  }

}

var _default = EntityAttributeValue;
exports.default = _default;
//# sourceMappingURL=entityAttributeValue.js.map