"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _ = require(".");

class EntityPool {
  constructor() {
    (0, _defineProperty2.default)(this, "pool", void 0);
    this.pool = {};
  }

  getPool() {
    return this.pool;
  }

  add(entity) {
    if (!this.getPool()[entity.constructor.name]) {
      this.getPool()[entity.constructor.name] = {};
    }

    this.getPool()[entity.constructor.name][entity.id] = new _.EntityPoolEntry(entity);
    return this;
  }

  has(entity, id) {
    const entityClass = entity.getClassName();

    if (!this.getPool()[entityClass]) {
      return false;
    }

    const entityId = _.Entity.isEntityInstance(entity) ? entity.id : id;
    return typeof this.getPool()[entityClass][entityId] !== "undefined";
  }

  remove(entity) {
    if (!this.getPool()[entity.constructor.name]) {
      return this;
    }

    delete this.getPool()[entity.constructor.name][entity.id];
    return this;
  }

  get(entity, id) {
    const entityClass = entity.getClassName();

    if (!this.getPool()[entityClass]) {
      return undefined;
    }

    const entityId = _.Entity.isEntityInstance(entity) ? entity.id : id;
    const poolEntry = this.getPool()[entityClass][entityId];

    if (poolEntry) {
      return poolEntry.getEntity();
    }

    return undefined;
  }

  flush() {
    this.pool = {};
    return this;
  }

}

var _default = EntityPool;
exports.default = _default;
//# sourceMappingURL=entityPool.js.map